import os

def install_package(package_name):
    package_name = package_name.replace("_", "-")  # 下载pip fake_useragent 包时  包名是:fake-useragent
    p = os.popen("pip list --format=columns")  # 获取所有包名 直接用 pip list 也可获取
    pip_list = p.read()  # 读取所有内容
    if package_name in pip_list:
        print("已经安装{}".format(package_name))
        return True
    else:
        print("没有安装{}!即将自动安装,请稍后".format(package_name))
        p = os.popen("pip install {}".format(package_name))
        if "Success" in p.read():
            print("安装{}成功!".format(package_name))
            return True if "Success" in p.read() else False


import os
# import json
from . import jsn as json

from hashlib import md5
from random import shuffle, randint

from typing import Any, Callable

def md5(string: str):
    from hashlib import md5 as HASH_FN
    return HASH_FN(string.encode()).hexdigest()

def hash_string(string: str, fn: Callable[[str, ], str]=None) -> str:
    """Generate unique ID for a string.
    """
    if fn is None:
        return md5(string.encode()).hexdigest()
    else:
        return fn(string)


## 对字典键值对的唯一性进行哈希
def to_str(data: Any) -> str:
    if isinstance(data, dict):
        ks = sorted(list(data.keys()))
        d = {k: to_str(data[k]) for k in ks}
        return json.dumps(d)
    else:
        return json.dumps(data)

def hash_dict_content(data: dict) -> str:
    s = to_str(data)
    return hash_string(s)

def random_id(length=64, use: str='NAa', chars=''):
    x = {
        'N': '1234567890',
        'a': 'abcdefghijklmnopqrstuvwxyz',
        'A': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
        'S': '~`!@#$%^&*()_-+=[]{}\|;:\'",.<>/?'
    }
    y = {'N': 10, 'a': 26, 'A': 26, 'S': 33}

    keys = []
    for k in x:
        if k in use:
            keys.append(k)
            chars += x[k]

    chars = [i for i in chars]
    shuffle(chars)

    seq = ''
    for i in range(length):
        seq += chars[randint(0, len(chars)-1)]

    return seq

## 通过链式字符串取出多层哈希的值
## 自动识别键的前n个唯一字符
def get_by_chain(hash, keys, sep='.'):
    """
    Get data from multi-level hash object.
    Args:
        `hash`: `dict` object whose elements are base types or new hash objects.
        `keys`: `str` object as key route where different-level keys are joint by dot.
                For example, `keys="xx.yy.zz"` means `hash['xx']['yy']['zz']`.
        `sep`: separate character used in keys chain, default is dot (`.`).
    Return: 
        data stored in `hash` by specific querying route.
    """
    key_list = keys.split(sep)
    val = hash
    for k in key_list:
        ks = list(val.keys())
        if k in ks:
            val = val[k]
        else:
            possible_ks = []
            for _ks in ks:
                if _ks.startswith(k):
                    possible_ks.append(_ks)
            if len(possible_ks) != 1:
                raise KeyError(f"Key `{k}` not in [{', '.join([str(i) for i in ks])}]")
            val = val[possible_ks[0]]
    return val


class Help:
    """
    Load help docment from multi-level JSON file.
    Args:
        `fp`: `str` object. JSON file, default is "help.json" in current directory.
        `language`: optional "zh" or "en", default is "zh", language of help text.
                    If not provided in the file, return empty string.
    Usage:
        Use `obj[KEYS_CHAIN]` to access data under specific route in `obj` which is
        a `Help` object related to specific JSON file. Especially, elementary key of
        KEYS_CHAIN joined by dot can be a shorted prefix string, as long as there is
        only one actual key starting with the shorter string. Otherwise it will raise
        a `KeyError`.
    """
    def __init__(self, fp=None, language='zh') -> None:
        if fp is None: fp = 'help.json'
        self.data = json.load(fp)
        self.language = language

    def __getitem__(self, index):
        """
        Get data using keys chain.
        Args:
            `index`: `str` object. Different-level keys joint by dot.
        Returns:
            Language-specific data according to the preset language.
        """
        d = get_by_chain(self.data, index)
        return d.get(self.language, '')


import os
import glob
from typing import Dict, Sequence

def get_versions(pattern, sep='-', num_prefix=2) -> Dict[str, Sequence[int]]:
    """获取指定目录下的版本数组
    Args:
        pattern: glob路径模板
        sep: 程序名与版本的分隔符
        num_prefix: 文件后缀数目
    Returns:
        多个版本号的数组，每个版本号由一组整数表示
    """
    versions = {}
    for fp in glob.glob(pattern):
        arr = os.path.basename(fp).split(sep)[1].split('.')[:-num_prefix]
        versions[fp] = [int(i) for i in arr]
        # versions.append([int(i) for i in arr])
    return versions

def get_latest_index(versions: Sequence[Sequence[int]]) -> int:
    """
    Args:
        versions: 多个版本号的数组，每个版本号由一组整数表示
    Return:
        最新版本在versions中的索引
    """
    max_index = len(versions[0]) - 1
    
    def _recursive(used: Sequence[int], i: int) -> int:
        """
        Args:
            used: 索引数组，在versions的子集里面查找
            i: 在每个version的第i个元素里面比较
        Return:
            最新版本在versions中的索引
        Exceptions:
            存在多个最新版本时会报错
        """
        maxv = -1
        _versions = []
        _index = []
        for vidx,version in enumerate(versions):
            if vidx not in used or version[i] < maxv:
                continue
            if version[i] > maxv:
                maxv = version[i]
                _versions = [version]
                _index = [vidx]
            elif version[i] == maxv:
                _versions.append(version)
                _index.append(vidx)

        if len(_versions) == 1:
            return _index[0]
        else:
            if i == max_index:
                raise Exception('Multiple latest versions!')
            return _recursive(_index, i + 1)

    return _recursive(list(range(len(versions))), 0)

def get_latest_file(pattern, *args, **kwargs):
    data = get_versions(pattern, *args, **kwargs)
    files = []
    versions = []
    for k, v in data.items():
        files.append(k)
        versions.append(v)

    max_index = get_latest_index(versions)
    max_file = files[max_index]

    return max_file


def auto_as(obj):
    if isinstance(obj, str):
        if obj.lower() == 'false':
            return False
        elif obj.lower() == 'true':
            return True
        else:
            try:
                _f = float(obj)
            except ValueError:
                # 不可转化为数值的字符串
                return obj
            else:
                try:
                    _i = int(obj)
                except ValueError:
                    # 可转化为浮点数但不能转化为整数的字符串
                    return _f
                else:
                    if _f == _i:
                        # 可转化为整数的字符串
                        return _i
                    else:
                        return _f

def to_multilines(text: str, ncols=20, eng=False):
    if eng:
        return text
    else:
        fmt_text = []
        index = 0
        while index < len(text):
            subtext = text[index:index+ncols]
            fmt_text.append(subtext)
            index += ncols
        fmt_text = '\n'.join(fmt_text)
        return fmt_text


if __name__ == '__main__':
    print(to_multilines('导入的项目名称，该名称将会直接作为项目名称保存到数据库。数据目录下必须存在以该名称命令的数据目录才能导入该项目的数据，请在导入前核对数据目录下的项目名称'))